program TutoGeneriques;

{$APPTYPE CONSOLE}

uses
  Types,
  SysUtils,
  Classes,
  TypInfo,
  Generics.Defaults,
  Generics.Collections,
  Generics.Trees in 'Generics.Trees.pas',
  Generics.CollectionsEx in 'Generics.CollectionsEx.pas',
  Generics.Nullable in 'Generics.Nullable.pas';

procedure WriteSquares(Max: Integer);
var
  List: TList<Integer>;
  I: Integer;
begin
  List := TList<Integer>.Create;
  try
    for I := 0 to Max do
      List.Add(I*I);

    for I := 0 to List.Count-1 do
      WriteLn(Format('%d*%0:d = %d', [I, List[I]]));
  finally
    List.Free;
  end;
end;

function DistanceToCenterSquare(const Point: TPoint): Integer; inline;
begin
  Result := Point.X*Point.X + Point.Y*Point.Y;
end;

function DistanceToCenter(const Point: TPoint): Extended; inline;
begin
  Result := Sqrt(DistanceToCenterSquare(Point));
end;

type
  TPointComparer = class(TComparer<TPoint>)
    function Compare(const Left, Right: TPoint): Integer; override;
  end;

function TPointComparer.Compare(const Left, Right: TPoint): Integer;
begin
  Result := DistanceToCenterSquare(Left) - DistanceToCenterSquare(Right);
end;

procedure SortPointsWithTPointComparer;
const
  MaxX = 100;
  MaxY = 100;
  PointCount = 10;
var
  List: TList<TPoint>;
  I: Integer;
  Item: TPoint;
begin
  List := TList<TPoint>.Create(TPointComparer.Create);
  try
    for I := 0 to PointCount-1 do
      List.Add(Point(Random(2*MaxX+1) - MaxX, Random(2*MaxY+1) - MaxY));

    List.Sort; // utilise le comparateur pass au constructeur

    for Item in List do
      WriteLn(Format('%d'#9'%d'#9'(distance au centre = %.2f)',
        [Item.X, Item.Y, DistanceToCenter(Item)]));
  finally
    List.Free;
  end;
end;

function ComparePoints(const Left, Right: TPoint): Integer;
begin
  Result := DistanceToCenterSquare(Left) - DistanceToCenterSquare(Right);
end;

procedure SortPointsWithComparePoints;
const
  MaxX = 100;
  MaxY = 100;
  PointCount = 10;
var
  List: TList<TPoint>;
  I: Integer;
  Item: TPoint;
begin
  List := TList<TPoint>.Create(TComparer<TPoint>.Construct(ComparePoints));
  try
    for I := 0 to PointCount-1 do
      List.Add(Point(Random(2*MaxX+1) - MaxX, Random(2*MaxY+1) - MaxY));

    List.Sort; // utilise le comparateur pass au constructeur

    for Item in List do
      WriteLn(Format('%d'#9'%d'#9'(distance au centre = %.2f)',
        [Item.X, Item.Y, DistanceToCenter(Item)]));
  finally
    List.Free;
  end;
end;

function MakeComparer(Reverse: Boolean = False): TComparison<TPoint>;
begin
  Result :=
    function(const Left, Right: TPoint): Integer
    begin
      Result := DistanceToCenterSquare(Left) - DistanceToCenterSquare(Right);
      if Reverse then
        Result := -Result;
    end;
end;

procedure SortPointsWithAnonymous;
const
  MaxX = 100;
  MaxY = 100;
  PointCount = 10;
var
  List: TList<TPoint>;
  I: Integer;
  Item: TPoint;
begin
  List := TList<TPoint>.Create(TComparer<TPoint>.Construct(
    MakeComparer(True)));

  try
    for I := 0 to PointCount-1 do
      List.Add(Point(Random(2*MaxX+1) - MaxX, Random(2*MaxY+1) - MaxY));

    List.Sort; // utilise le comparateur pass au constructeur

    for Item in List do
      WriteLn(Format('%d'#9'%d'#9'(distance au centre = %.2f)',
        [Item.X, Item.Y, DistanceToCenter(Item)]));
  finally
    List.Free;
  end;
end;

procedure TestGenericTree;
const
  NodeCount = 20;
var
  Tree, Node: TTreeNode<Integer>;
  I, J, MaxDepth, Depth: Integer;
begin
  Tree := TTreeNode<Integer>.Create(0);
  try
    // Crer l'arbre
    MaxDepth := 0;
    for I := 1 to NodeCount do
    begin
      Depth := Random(MaxDepth+1);
      Node := Tree;

      for J := 0 to Depth-1 do
      begin
        if Node.IsLeaf then
          Break;
        Node := Node.Children[Random(Node.ChildCount)];
      end;

      if TTreeNode<Integer>.Create(Node, I).Depth > MaxDepth then
        Inc(MaxDepth);
    end;

    // Afficher l'arbre avec un parcours prfixe
    Tree.PreOrderWalk(
      procedure(const Node: TTreeNode<Integer>)
      begin
        Write(StringOfChar(' ', 2*Node.Depth));
        Write('- ');
        WriteLn(Node.Value);
      end);
  finally
    Tree.Free;
  end;
end;

type
  TArrayUtils = class
  public
    class function Min<T>(const Items: array of T;
      const Comparer: IComparer<T>): T; overload; static;
    class function Min<T>(const Items: array of T;
      const Comparison: TComparison<T>): T; overload; static;
    class function Min<T: IComparable<T>>(
      const Items: array of T): T; overload; static;

    class function MinDefault<T>(
      const Items: array of T): T; static;
  end;

class function TArrayUtils.Min<T>(const Items: array of T;
  const Comparer: IComparer<T>): T;
var
  I: Integer;
begin
  if Length(Items) = 0 then
    raise Exception.Create('No items in the array');

  Result := Items[Low(Items)];
  for I := Low(Items)+1 to High(Items) do
    if Comparer.Compare(Items[I], Result) < 0 then
      Result := Items[I];
end;

class function TArrayUtils.Min<T>(const Items: array of T;
  const Comparison: TComparison<T>): T;
begin
  Result := Min<T>(Items, TComparer<T>.Construct(Comparison));
end;

class function TArrayUtils.Min<T>(const Items: array of T): T;
var
  Comparison: TComparison<T>;
begin
  Comparison :=
    function(const Left, Right: T): Integer
    begin
      Result := Left.CompareTo(Right);
    end;
  Result := Min<T>(Items, Comparison);
end;

class function TArrayUtils.MinDefault<T>(const Items: array of T): T;
begin
  Result := Min<T>(Items, IComparer<T>(TComparer<T>.Default));
end;

type
  TRTTI = class(TObject)
  public
    class procedure PrintType<T>; static;
    class function TypeInfo<T>: PTypeInfo; static;
  end;

class procedure TRTTI.PrintType<T>;
var
  Info: PTypeInfo;
begin
  Info := System.TypeInfo(T); // attention ! Info peut valoir nil ici

  if Info = nil then
    WriteLn('Ce type ne possde pas de RTTI')
  else
    WriteLn(Info.Name, #9,
      GetEnumName(System.TypeInfo(TTypeKind), Byte(Info.Kind)));
end;

class function TRTTI.TypeInfo<T>: PTypeInfo;
begin
  Result := System.TypeInfo(T);
end;

{var
  Max: Integer;}
var
  IntArray: array[0..8] of Integer = (5, 40, 2, 0, -4, -2, 3, -10, 4);
  Info: PTypeInfo;
  Source, Dest: TList<Integer>;
  Value: Integer;
  NullValue: TNullable<Integer>;
begin
  try
    Randomize;

    SortPointsWithTPointComparer;
    WriteLn;
    SortPointsWithComparePoints;
    WriteLn;
    SortPointsWithAnonymous;

    WriteLn;
    TestGenericTree;

    WriteLn;
    WriteLn(TArrayUtils.MinDefault<Integer>(IntArray));
    WriteLn(TArrayUtils.MinDefault<Integer>([5, 40, 2, 0, -4, -2, 3, -10, 4]));

    WriteLn;
    TRTTI.PrintType<Integer>;
    TRTTI.PrintType<TObject>;
    TRTTI.PrintType<Pointer>;
    TRTTI.PrintType<TComparison<Integer>>;
    TRTTI.PrintType<TTreeNode<Integer>>;

    Info := TRTTI.TypeInfo<Pointer>;
    WriteLn(Integer(Info));
    Info := TRTTI.TypeInfo<Integer>;
    WriteLn(Integer(Info));

    WriteLn;
    Source := TList<Integer>.Create;
    try
      Source.AddRange([2, -9, -5, 50, 4, -3, 7]);
      Dest := TList<Integer>.Create;
      try
        TListEx.FindAll<Integer>(Source, Dest, TPredicate<Integer>(
          function(const Value: Integer): Boolean
          begin
            Result := Value > 0;
          end));

        for Value in Dest do
          WriteLn(Value);
      finally
        Dest.Free;
      end;
    finally
      Source.Free;
    end;

    WriteLn;
    NullValue := 5;
    WriteLn(Integer(NullValue));
    NullValue := nil;
    if NullValue.IsNil then
      WriteLn('nil')
    else
      WriteLn(NullValue.Value);

    NullValue := 10;
    Value := NullValue;
    WriteLn(Value);

    {WriteLn('Entrez un nombre naturel :');
    ReadLn(Max);
    WriteSquares(Max);}
  except
    on E:Exception do
      Writeln(E.Classname, ': ', E.Message);
  end;

  ReadLn;
end.

